/*
 * Copyright 1997, Regents of the University of Minnesota
 *
 * par_metis.h
 *
 * This file includes all necessary header files
 *
 * Started 8/27/94
 * George
 *
 * $Id: parmetislib.h 10379 2011-06-22 18:51:24Z benjamin $
 */

/*
#define DEBUG			1
#define DMALLOC			1
*/

#include <GKlib.h>

#include <parmetis.h>

#include "../metis/libmetis/gklib_defs.h"

#include <mpi.h> 

#include <rename.h>
#include <defs.h>
#include <struct.h>
#include <macros.h>
#include <proto.h>

